#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
"""
wisestork.utils
~~~~~~~~~~~~~

:copyright: (c) 2016-2019 Sander Bollen
:license: GPL-3.0
"""


from collections import namedtuple

Bin = namedtuple("Bin", ["start", "end"])


def utf8(value):
    if isinstance(value, bytes):
        return value
    if not isinstance(value, str):
        value = str(value)
    return bytes(value, "utf-8")


def as_str(value):
    if isinstance(value, bytes):
        return value.decode()
    return value


class BedLine(namedtuple("BedLine", ["chromosome", "start", "end", "value"])):
    __slots__ = ()

    def __str__(self):
        return "{0}\t{1}\t{2}\t{3}".format(as_str(self.chromosome),
                                           self.start, self.end,
                                           as_str(self.value))

    def __bytes__(self):
        return b"\t".join(map(utf8, [getattr(self, x) for x in self._fields]))

    def __eq__(self, other):
        return self.chromosome == other.chromosome and \
               self.start == other.start and self.end == other.end

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def fromline(cls, line, sep=b"\t"):
        contents = utf8(line).split(sep)
        if len(contents) == 3:
            return cls(*map(attempt_numeric, contents), value="NA")
        else:
            return cls(*map(attempt_numeric, contents))


def attempt_numeric(value):
    """
    Attempt to make an numeric from a string
    first try to make an integer
    When that is not possible, make a float.
    When that too fails, return the string
    :param value: the input string
    :return: the possibly-converted item
    """
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value


def get_bins(chromosome_length, binsize):
    """
    Get list of 2-tuples of start and end positions of bins for a given
    chromosome
    :param chromosome_length: integer
    :param binsize: integer
    :return: list of 2-tuples of (start, end). Start = 0-based
    """

    starts = list(range(0, chromosome_length, binsize))
    ends = []
    for s in starts:
        if s + binsize < chromosome_length:
            ends.append(s+binsize)
        else:
            ends.append(chromosome_length)
    return [Bin(s, e) for s, e in zip(starts, ends)]


class BedReader(object):
    """
    Iterator for reading bed files
    Returns instances of BedLine

    If no value is found in the 4th column, val = 'NA'
    """

    def __init__(self, filename):
        self.filename = filename
        self.__handle = open(self.filename, 'rb')

    def __iter__(self):
        return self

    def __next__(self):
        try:
            line = self.__handle.readline()
        except StopIteration:
            self.__handle.close()
            raise StopIteration
        contents = line.strip().split(b"\t")
        if len(contents) == 3:
            return BedLine(*map(attempt_numeric, contents), value='NA')
        elif len(contents) == 4:
            return BedLine(*map(attempt_numeric, contents))
        else:
            self.__handle.close()
            raise StopIteration

    def next(self):
        return self.__next__()



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
import pytest
from wisestork.utils import (BedLine, utf8, as_str, attempt_numeric,
                             get_bins, BedReader)


@pytest.fixture
def test_bedline():
    return BedLine(b"chr1", 1, 100, 20)


class TestFunctions:

    def test_utf8(self):
        assert utf8(b"test") == b"test"
        assert utf8("test") == b"test"
        assert utf8(1) == b"1"

    def test_as_str(self):
        assert as_str("test") == "test"
        assert as_str(b"test") == "test"

    def test_attempt_numeric(self):
        assert attempt_numeric("2") == 2
        assert attempt_numeric(b"2") == 2
        assert attempt_numeric("2.5") == 2.5
        assert attempt_numeric(b"2.5") == 2.5
        assert attempt_numeric("test") == "test"
        assert attempt_numeric(b"test") == b"test"

    def test_get_bins(self):
        test = get_bins(500, 100)
        assert len(test) == 5
        assert test[0].start == 0
        assert test[0].end == 100
        assert test[1].start == 100
        assert test[1].end == 200
        assert test[4].start == 400
        assert test[4].end == 500


class TestBedline(object):

    def test_str(self, test_bedline):
        assert str(test_bedline) == "chr1\t1\t100\t20"

    def test_bytes(self, test_bedline):
        assert bytes(test_bedline) == b"chr1\t1\t100\t20"

    def test_eq(self, test_bedline):
        assert test_bedline == test_bedline

    def test_neq(self, test_bedline):
        other_line = BedLine(b"chr2", 1, 100, 20)
        assert test_bedline != other_line

    def test_fromline(self, test_bedline):
        n = BedLine.fromline("chr1\t1\t100\t20")
        assert n == test_bedline
        n2 = BedLine.fromline(b"chr1\t1\t100\t20")
        assert n2 == test_bedline
        three_fields = BedLine.fromline("chr1\t1\t100")
        assert three_fields == BedLine(b"chr1", 1, 100, b"NA")


class TestBedReader:

    def test_bed(self):
        reader = BedReader(filename="test/data/test.bed")
        records = [x for x in reader]
        assert len(records) == 5
        assert all([x.value == "NA" for x in records])
        assert all([x.chromosome == b"chr1" for x in records])

    def test_bedgraph(self):
        reader = BedReader(filename="test/data/test.bedgraph")
        records = [x for x in reader]
        assert len(records) == 5
        assert records[0].value == 10
        assert records[-1].value == 40
        assert all([x.chromosome == b"chr1" for x in records])



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
from pyfaidx import Fasta
from pytest import fixture

from wisestork.gc_correct import filter_bin, correct
from wisestork.utils import BedLine, attempt_numeric


@fixture
def bedlines():
    with open("test/data/count.bed") as count_handle:
        bedlines = [BedLine(*map(attempt_numeric, x.split("\t"))) for x in
                    count_handle]
    return bedlines


@fixture
def fasta():
    return Fasta("test/data/chrQ.fasta")


class TestFunctions:
    def test_filter_bin(self, bedlines, fasta):
        assert all([filter_bin(x, fasta, 0.1, 0.001) for x in bedlines])
        assert all([not filter_bin(x, fasta, 0.1, 5) for x in bedlines])

    def test_correct(self, bedlines, fasta):
        assert all([x.value == 0 for x in correct(bedlines, fasta, frac_r=5)])
        corrected = correct(bedlines, fasta)
        for i in range(4):
            assert 0.9 < corrected[i].value < 1.1
        assert 0.4 < corrected[4].value < 0.6



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
from pyfaidx import Fasta

from wisestork.gc import get_n_per_bin, get_gc_for_bin
from wisestork.utils import get_bins


class TestFunctions:

    def test_n_per_bin(self):
        fa = Fasta("test/data/chrQ.fasta")
        bins = get_bins(500, 100)
        for bin in bins:
            assert get_n_per_bin(fa, 'chrQ', bin) == 0

    def test_get_gc_for_bin(self):
        fa = Fasta("test/data/chrQ.fasta")
        bins = get_bins(500, 100)
        assert get_gc_for_bin(fa, "chrQ", bins[0]) == 47
        assert get_gc_for_bin(fa, "chrQ", bins[1]) == 50
        assert get_gc_for_bin(fa, "chrQ", bins[2]) == 54
        assert get_gc_for_bin(fa, "chrQ", bins[3]) == 58
        assert get_gc_for_bin(fa, "chrQ", bins[4]) == 50



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
"""
wisestork.count
~~~~~~~~~~~~~~
:copyright: (c) 2016-2019 Sander Bollen
:license: GPL-3.0
"""
import pysam

from .utils import get_bins, BedReader, BedLine


def reads_per_bin(bam_reader, chromosome, bin):
    """
    Number of reads per bin
    :param bam_reader: an instance of pysam.AlignmentFile
    :param chromosome: chromosome name
    :param bin: Bin namedtuple
    :return: integer
    """
    try:
        reads = bam_reader.count(chromosome, bin.start, bin.end)
    except ValueError:
        reads = 0

    return reads


def get_chromosomes_from_header(header):
    """
    Get list of chromosome names from bam headers
    :param header: bam header
    :return: list of tuples of (chromosome names, chromosome_length)
    """
    return [(x['SN'], x['LN']) for x in header['SQ']]


def get_chromosomes_from_fasta(fa):
    """
    Get chromosome names from fasta file
    :param fa: instance of pyfaidx.Fasta
    :return: list of tuples of (chromosome names, chromosome_lenght)
    """
    return [(x, len(fa[x])) for x in fa.keys()]


def count(input, output, binsize, reference, binfile=None):
    """
    Main function for counting reads per bin
    :param input: Path to input BAM
    :param output: path to output BED
    :param binsize: binsize
    :param reference: path to reference fasta
    """
    # TODO: check whether chromosome names in reference match those in bam file
    samfile = pysam.AlignmentFile(input, 'rb')
    chromosomes = get_chromosomes_from_header(samfile.header)
    with open(output, "wb") as ohandle:
        if binfile:
            bin_reader = BedReader(binfile)
            for bin in bin_reader:
                val = reads_per_bin(samfile, bin.chromosome, bin)
                bed = BedLine(bin.chromosome, bin.start, bin.end, val)
                ohandle.write(bytes(bed) + b"\n")
        else:
            for ch, ln in chromosomes:
                for bin in get_bins(ln, binsize):
                    val = reads_per_bin(samfile, ch, bin)
                    bed = BedLine(ch, bin.start, bin.end, val)
                    ohandle.write(bytes(bed) + b"\n")



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
import pysam
import pyfaidx

from tempfile import NamedTemporaryFile
from hashlib import sha1
from wisestork.count import (get_chromosomes_from_header, reads_per_bin,
                             get_chromosomes_from_fasta, get_bins, count)


class TestFunctions:

    def test_sam_header(self):
        sam = pysam.AlignmentFile("test/data/test.bam")
        chroms = get_chromosomes_from_header(sam.header)
        assert len(chroms) == 1
        assert chroms[0][0] == "chrQ"
        assert chroms[0][1] == 500

    def test_fq_header(self):
        fa = pyfaidx.Fasta("test/data/chrQ.fasta")
        chroms = get_chromosomes_from_fasta(fa)
        assert len(chroms) == 1
        assert chroms[0][0] == "chrQ"
        assert chroms[0][1] == 500

    def test_count(self):
        sam = pysam.AlignmentFile("test/data/test.bam")
        bins = get_bins(500, 100)
        assert reads_per_bin(sam, "chrQ", bins[0]) == 40
        assert reads_per_bin(sam, "chrQ", bins[1]) == 52
        assert reads_per_bin(sam, "chrQ", bins[2]) == 34
        assert reads_per_bin(sam, "chrQ", bins[3]) == 42
        assert reads_per_bin(sam, "chrQ", bins[4]) == 26
        assert reads_per_bin(sam, "NotExist", bins[0]) == 0


class TestMain:

    def test_main(self):
        tmp_file = NamedTemporaryFile()
        count("test/data/test.bam", tmp_file.name, 100, "test/data/chrQ.fasta")
        output = b"".join(open(tmp_file.name, "rb").readlines())
        expected = b"".join(open("test/data/count.bed", "rb").readlines())
        assert sha1(output).hexdigest() == sha1(expected).hexdigest()

    def test_with_binfile(self):
        tmp_file = NamedTemporaryFile()
        count("test/data/test.bam", tmp_file.name, 100,
              "test/data/chrQ.fasta", "test/data/regions.bed")
        output = b"".join(open(tmp_file.name, "rb").readlines())
        expected = b"".join(open("test/data/count.bed", "rb").readlines())
        assert sha1(output).hexdigest() == sha1(expected).hexdigest()



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.

from click.testing import CliRunner
import pytest

from wisestork.wisestork import count_cli, gcc_cli, newref_cli, zscore_cli


@pytest.fixture()
def runner():
    return CliRunner()


def test_cli_count_help(runner):
    result = runner.invoke(count_cli, "--help")
    assert result.exit_code == 0


def test_cli_gc_correct_help(runner):
    result = runner.invoke(gcc_cli, "--help")
    assert result.exit_code == 0


def test_cli_newref_help(runner):
    result = runner.invoke(newref_cli, "--help")
    assert result.exit_code == 0


def test_cli_zscore_help(runner):
    result = runner.invoke(zscore_cli, "--help")
    assert result.exit_code == 0



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
"""
wisestork.gc_correct
:copyright: (c) 2016-2019 Sander Bollen
:license: GPL-3.0
"""

import numpy as np
import warnings

import statsmodels.nonparametric.smoothers_lowess as statlow
from pyfaidx import Fasta

from .utils import BedLine, attempt_numeric
from .gc import get_gc_for_bin, get_n_per_bin


def filter_bin(bin, ref_fasta, frac_n, frac_r):
    """
    Return True if a 'correct' bin, return False if an 'incorrect' bin
    :param bin: BedLine namedtuple
    :param ref_fasta: an instance of pyfaidx.Fasta
    :param frac_n: maximal fraction of N-bases per bin
    :param frac_r: minimum fraction of reads per bin
    :return: Boolean
    """
    ns = get_n_per_bin(ref_fasta, bin.chromosome, bin)
    return ns < ((bin.end - bin.start)*frac_n) and bin.value > ((bin.end - bin.start)*frac_r)  # noqa


def correct(inputs, fasta, frac_n=0.1, frac_r=0.0001, lowess_iter=3,
            lowess_frac=0.1):
    """
    GC-correct input bed lines.
    GC correction takes place with a local regression (LOWESS) on GC perc vs
    number of reads
    :param inputs: list of BedLine namedtuples
    :param fasta: instance of pyfaidx.Fasta
    :param frac_n: maximal fraction on N-bases per bin
    :param frac_r: minimum fraction of reads per bin
    :param lowess_iter: amount of iterations of LOWESS function
    :param lowess_frac: fraction of input data used for LOWESS function
    :return: corrected BedLines
    """
    reads = []
    gcs = []
    for line in inputs:
        if filter_bin(line, fasta, frac_n, frac_r):
            gcs.append(get_gc_for_bin(fasta, line.chromosome, line))
            reads.append(line.value)

    reads = np.array(reads, np.float)
    gcs = np.array(gcs, np.float)
    if lowess_frac*len(reads) < 4 and len(reads) > 0:
        # need at least four data ponts
        warnings.warn("Too few data points for lowess. Raising lowess_frac")
        lowess_frac = 4.0/len(reads)
        delta = 0  # remove delta in this case
    else:
        delta = 0.01 * len(gcs)
    lowess = statlow.lowess(reads, gcs, return_sorted=False,
                            delta=delta, frac=lowess_frac,
                            it=lowess_iter).tolist()

    corrected_lines = []

    for line in inputs:
        if filter_bin(line, fasta, frac_n, frac_r):
            corr_val = float(line.value) / lowess.pop(0)
        else:
            corr_val = 0
        n_bed = BedLine(line.chromosome, line.start, line.end, corr_val)
        corrected_lines.append(n_bed)

    return corrected_lines


def gc_correct(input, output, reference, frac_n, frac_r, iter, frac_lowess):
    fasta = Fasta(reference)
    bed_lines = [BedLine(*map(attempt_numeric, x.split("\t"))) for
                 x in open(input)]
    corrected = correct(bed_lines, fasta, frac_n, frac_r, iter, frac_lowess)

    with open(output, "wb") as ohandle:
        for line in corrected:
            ohandle.write(bytes(str(line) + "\n", 'utf-8'))



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.

"""
wisestork.wisestork
~~~~~~~~~~~~~~~~~~~~~
:copyright: (c) 2016-2019 Sander Bollen
:license: GPL-3.0
"""

import click

from .count import count
from .gc_correct import gc_correct
from .newref import newref
from .ztest import ztest
from . import version as wiseguy_version

shared_options = [
    click.option("--binsize", "-B", type=click.IntRange(0, None),
                 default=50000,
                 help="Bin size to use. Default = 50000"),
    click.option("--reference", "-R", type=click.Path(exists=True),
                 required=True,
                 help="Path to reference fasta"),
    click.option('--bin-file', '-L', type=click.Path(exists=True),
                 required=False,
                 help="Optional path to region BED file"),
    click.version_option(version=wiseguy_version())
]


def generic_option(options):
    """
    Decorator to add generic options to Click CLI's
    The group parent should NOT be decorated with this decorator
    :param options: list of click.option
    :return: decorated function
    """
    def __generic_option(func):
        for option in reversed(options):
            func = option(func)
        return func
    return __generic_option


@click.command(short_help="Count coverages")
@generic_option(shared_options)
@click.option("--output", "-O", type=click.Path(), required=True,
              help="Path to output BED file")
@click.option("--input", "-I", type=click.Path(exists=True), required=True,
              help="Path to input BAM file")
def count_cli(**kwargs):
    """
    Take a BAM file, and calculate the number of reads per bin.
    It will output a BED file (with 0-based positions) of regions and
    associated number of reads per bin.

    \b
    Your BAM file _must_ be indexed and _must_ contain chromosome
    lengths and names in the header.
    """
    input = kwargs.get("input", None)
    output = kwargs.get("output", None)
    binsize = kwargs.get("binsize", 50000)
    reference = kwargs.get("reference", None)
    regions = kwargs.get("bin_file", None)
    count(input=input, output=output, binsize=binsize, reference=reference,
          binfile=regions)


@click.command(short_help="GC correct")
@generic_option(shared_options)
@click.option("--output", "-O", type=click.Path(), required=True,
              help="Path to output BED file")
@click.option("--input", "-I", type=click.Path(exists=True), required=True,
              help="Path to input BED file")
@click.option("--frac-n", "-n", type=click.FLOAT, default=0.1,
              help="Maximum fraction of N-bases per bin. Default = 0.1")
@click.option("--frac-r", "-r", type=click.FLOAT, default=0.0001,
              help="Minimum fraction of reads per bin. Default = 0.0001")
@click.option("--iter", "-t", type=click.INT, default=3,
              help="Number of iterations for LOWESS function. Default = 3")
@click.option("--frac-lowess", "-l", type=click.FLOAT, default=0.1,
              help="Fraction of data to use for LOWESS function. "
                   "Default = 0.1")
def gcc_cli(**kwargs):
    """
    GC-correct a BED file containing counts per region
    """
    input_path = kwargs.get("input", None)
    output = kwargs.get("output", None)
    reference = kwargs.get("reference", None)
    frac_n = kwargs.get("frac_n", 0.1)
    frac_r = kwargs.get("frac_r", 0.0001)
    iter = kwargs.get("iter", 3)
    frac_lowess = kwargs.get("frac_lowess", 0.1)
    gc_correct(input=input_path, output=output, reference=reference,
               frac_r=frac_r, frac_n=frac_n, iter=iter,
               frac_lowess=frac_lowess)


@click.command(short_help="Calculate Z-scores")
@generic_option(shared_options)
@click.option("--input", "-I", type=click.Path(exists=True), required=True,
              help="Path to input BED file")
@click.option("--output", "-O", type=click.Path(), required=True,
              help="Path to output BED file")
@click.option("--dictionary-file", "-D", type=click.Path(), required=True,
              help="Path to dictionary BED file")
def zscore_cli(**kwargs):
    """
    Calculate Z-scores from GC-corrected BED files.

    \b
    You must supply a "reference dictionary" BED file
    containing locations of reference bins.
    This reference dictionary must be gzipped and
    indexed with tabix.

    \b
    Your query BED file should also be gzipped and
    indexed with tabix
    """
    input_path = kwargs.get("input", None)
    output_path = kwargs.get("output", None)
    database = kwargs.get("dictionary_file", None)
    ztest(input_path=input_path, output_path=output_path,
          database_path=database)


@click.command(short_help="Create new reference")
@generic_option(shared_options)
@click.option("--input", "-I", type=click.Path(exists=True), required=True,
              multiple=True, help="Path(s) to input BEDs")
@click.option("--output", "-O", type=click.Path(), required=True,
              help="Path to output BED file")
@click.option("--n-bins", "-n", type=click.INT, default=250,
              help="Amount of neighbours bins to consider per bin")
def newref_cli(**kwargs):
    """
    Create a new reference dictionary BED file.

    \b
    This tool takes multiple input BED files.
    For each bin, it will then find a set of neighbour bins
    that behave most similar to one another.

    \b
    You must short and then tabix the output after running this tool.
    """
    input_path = kwargs.get("input", None)
    output_path = kwargs.get("output", None)
    reference_fasta = kwargs.get("reference", None)
    binsize = kwargs.get("binsize", 50000)
    n_bins = kwargs.get("n_bins", 250)
    regions = kwargs.get("bin_file", None)
    newref(input_paths=input_path, output_path=output_path,
           reference=reference_fasta,
           binsize=binsize, n_bins=n_bins, binfile=regions)


@click.group()
@click.version_option()
def cli(**kwargs):
    """
    Discover CNVs from BAM files.

    \b
    A typical workflow first extracts regions from a BAM file
    The resulting BED tracks must then be GC-corrected.
    Using a reference track of region similarity,
    One can then calculate Z-scores for every region.

    \b
    The following sub-commands are supported:
     - count: count coverage per bin
     - gc-correct: GC-correct bins
     - zscore: calculate Z-scores
     - newref: Generate a new reference dictionary of bin similarities

    """
    pass


def main():
    cli.add_command(count_cli, "count")
    cli.add_command(gcc_cli, "gc-correct")
    cli.add_command(zscore_cli, "zscore")
    cli.add_command(newref_cli, "newref")
    cli()


if __name__ == '__main__':
    main()



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
from collections import namedtuple
import random
from math import isnan

from wisestork.ztest import create_key,  get_z_score
from wisestork.utils import BedLine

ValueObject = namedtuple("ValueObject", ['value'])  # little helper object


class TestFunctions:

    def test_create_key(self):
        bline = BedLine(b"chr1", 1, 100, 0)
        bline2 = BedLine(b"chr1", b"1", b"100", 0)
        assert create_key(bline) == b"chr1|1|100"
        assert create_key(bline2) == b"chr1|1|100"

    def test_build_reference_index(self):
        pass

    def test_get_z_score(self):
        assert isnan(get_z_score(None, []))
        objects = [ValueObject(random.normalvariate(100, 20)) for _ in
                   range(2000)]
        zero_test = ValueObject(100)
        assert -0.5 < get_z_score(zero_test, objects) < 0.5
        minus_test = ValueObject(20)
        assert -4.5 < get_z_score(minus_test, objects) < -3.5
        positive_test = ValueObject(160)
        assert 2.5 < get_z_score(positive_test, objects) < 3.5
        zero_std = [ValueObject(random.normalvariate(100, 0)) for _ in
                    range(2000)]
        assert isnan(get_z_score(zero_test, zero_std))

    def test_ztest(self):
        pass



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.

"""
wisestork.newref
~~~~~~~~~~~~~~
:copyright: (c) 2016-2019 Sander Bollen
:license: GPL-3.0
"""

import math
import numpy as np

from .utils import BedLine, get_bins, BedReader
from pyfaidx import Fasta


def get_unique_bins(fasta, binsize):
    """
    Get a list of unique bins (by position, not value!)
    :param fasta: pyfaidx.Fasta instance
    :return: bins
    """
    bins = []
    for contig in fasta:
        n = contig.name
        tmp_bins = get_bins(len(contig), binsize)
        bins += [BedLine(n, x[0], x[1], 0) for x in tmp_bins]

    return bins


def build_main_list(bins, binsize, reference, binfile=None):
    """
    Build main list of bins.
    All unique positions are selected
    Then the median value of all bins on those positions are calculated
    :param bins: list of bins
        order of this list should correspond to input files
        e.g. if 3 bins per file, this list looks like
        [1,2,3,1,2,3,1,2,3 ... ]
    :param binsize: binsize:
    :param reference: instance of pyfaidx.Fasta
    :return: list of bins (1 per position), sorted by median value
    """
    if not binfile:
        unique_positions = get_unique_bins(reference, binsize)
    else:
        reader = BedReader(binfile)
        unique_positions = [x for x in reader]
    median_bins = []
    for i, u in enumerate(unique_positions):
        working_bins = bins[i::len(unique_positions)]
        med = np.median([x.value for x in working_bins])
        median_bins.append(BedLine(u.chromosome, u.start, u.end, med))

    return sorted(median_bins, key=lambda x: x.value)


class ReferenceBinGenerator(object):
    """
    Iterator for generating reference bins
    Every item will be 2-tuple of (Bin, List[Bins])
    With the second item being similar bins
    """

    def __init__(self, inputs, n_bins, reference, binsize=int(1e6),
                 binfile=None):
        """
        Create instance of ReferenceBinGenerator
        :param inputs: list of paths to files of gc-corrected bedgraph files
        :param n_bins: number of neighbour bins to consider
        """
        self.inputs = inputs
        self.n_bins = n_bins
        self.fasta = Fasta(reference)
        self.binsize = binsize
        self.binfile = binfile
        self.__bins = self.get_all_bins()
        self.__idx = 0

    def get_all_bins(self):
        bins = []
        for inp in self.inputs:
            tmp_reader = BedReader(inp)
            bins += [x for x in tmp_reader]
        return build_main_list(bins, self.binsize, self.fasta, self.binfile)

    def __next__(self):
        if self.__idx == len(self.__bins):
            raise StopIteration
        nearest = self.get_nearest_at_idx(self.__idx)
        cur = self.__bins[self.__idx]
        filtered = self.filter_bins(cur, nearest)
        self.__idx += 1
        return cur, filtered

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self

    def get_nearest_at_idx(self, idx):
        e_dist = len(self.__bins) - idx
        # starting edge; if distance from start less
        # or equal to half the window size
        if idx == 0 or 0 < idx <= self.n_bins//2:
            vals = self.__bins[:self.n_bins]
        # ending edge; if distance from end is less
        # or equal to half the window size
        elif e_dist <= self.n_bins//2:
            vals = self.__bins[-self.n_bins:]
        else:
            vals = self.__bins[idx-(self.n_bins//2):idx+int(math.ceil((self.n_bins/2)))]  # noqa
        return vals

    def filter_bins(self, target_bin, bins):
        """
        Filter a list of reference bins
        Bins to exclude:
         * target bin
         * Bins with value > 3x mean+stdev
         * Adjacent bins # TODO
        :param target_bin:
        :param bins:
        :return: filtered list of bins (may be empty)
        """
        ref_bins = [x for x in bins if x != target_bin]
        if len(ref_bins) == 0:
            return ref_bins

        stdev = np.std([x.value for x in ref_bins])
        mean = np.mean([x.value for x in ref_bins])
        non_outliers = [x for x in ref_bins if mean+(3*stdev) > x.value > mean-(3*stdev)]  # noqa
        return non_outliers


def newref(input_paths, output_path, reference, binsize, n_bins=250,
           binfile=None):
    """
    Create a new reference bed file
    :param input_paths: paths to gc-corrected BED files
    :param output_path: path to output bed file
    :param reference: path to reference Fasta
    :param binsize: binsize
    :param n_bins: number of neighbour bins to consider
    """
    gen = ReferenceBinGenerator(input_paths, n_bins, reference, binsize,
                                binfile)
    ohandle = open(output_path, "wb")
    for x in gen:
        chrom = x[0].chromosome
        start = x[0].start
        end = x[0].end
        references = []
        for record in x[1]:
            fmt = "{0},{1},{2}"
            if isinstance(record.chromosome, str):
                references.append(fmt.format(record.chromosome,
                                             record.start, record.end))
            else:
                references.append(fmt.format(record.chromosome.decode(),
                                             record.start, record.end))
        references = "|".join(references)

        if len(references) == 0:
            references = np.nan

        t = BedLine(chrom, start, end, references)
        ohandle.write(bytes(t) + b"\n")



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
import hashlib
from os import remove
from tempfile import NamedTemporaryFile
import pytest
from pyfaidx import Fasta

from wisestork.newref import (build_main_list, get_unique_bins,
                              ReferenceBinGenerator, newref)
from wisestork.utils import BedReader, BedLine


@pytest.fixture(scope="module")
def fuzzed_files():
    multipliers = [1.5, 2.0, 3.0, 3.5, 4.0]
    init = [x for x in BedReader("test/data/gc_correct.bed")]
    fs = []
    for multi in multipliers:
        tmp = NamedTemporaryFile(delete=False)
        t = [BedLine(x.chromosome, x.start, x.end, x.value*multi) for x in
             init]
        for z in t:
            tmp.write(bytes(z) + b"\n")
        tmp.close()
        fs.append(tmp)

    yield [x.name for x in fs]
    # teardown after this
    for f in fs:
        remove(f.name)


@pytest.fixture(scope="module")
def input_bins(fuzzed_files):
    inp = []
    for z in fuzzed_files:
        inp += [x for x in BedReader(z)]
    return inp


@pytest.fixture
def fasta():
    return Fasta("test/data/chrQ.fasta")


class TestFunctions:

    def test_get_unique_bins(self, fasta):
        bins = get_unique_bins(fasta, 100)
        assert len(bins) == 5
        for i, x in enumerate(bins):
            assert x.start == i*100
            assert x.end == (i+1)*100

    def test_build_main_list(self, input_bins, fasta):
        main = build_main_list(input_bins, 100, fasta)
        assert len(main) == 5
        assert main[0].start == 400
        assert main[1].start == 100
        assert main[2].start == 0
        assert main[3].start == 200
        assert main[4].start == 300
        main_w_f = build_main_list(input_bins, 100, fasta,
                                   "test/data/regions.bed")
        assert [x.start for x in main_w_f] == [x.start for x in main]


class TestReferenceBinGenerator:

    def test_init(self, fuzzed_files, fasta):
        ref1 = ReferenceBinGenerator(fuzzed_files, 5, fasta.filename, 100)
        assert len(ref1.get_all_bins()) == 5
        ref2 = ReferenceBinGenerator(fuzzed_files, 5, fasta.filename, 100,
                                     "test/data/regions.bed")
        assert len(ref2.get_all_bins()) == 5
        assert [x.start for x in ref1.get_all_bins()] == [x.start for x in
                                                          ref2.get_all_bins()]

    def test_get_nearest_and_filter(self, fuzzed_files, fasta):
        ref1 = ReferenceBinGenerator(fuzzed_files, 5, fasta.filename, 100)
        for i in range(5):
            a = ref1.get_nearest_at_idx(i)
            assert len(a) == 5
            filt = ref1.filter_bins(ref1.get_all_bins()[i], a)
            assert len(filt) == 4
        ref2 = ReferenceBinGenerator(fuzzed_files, 3, fasta.filename, 100)
        for i in range(5):
            a = ref2.get_nearest_at_idx(i)
            assert len(a) == 3
            filt = ref2.filter_bins(ref2.get_all_bins()[i], a)
            assert len(filt) == 2

    def test_iteration(self, fuzzed_files, fasta):
        ref1 = ReferenceBinGenerator(fuzzed_files, 5, fasta.filename, 100)
        refs = [x for x in ref1]
        assert len(refs) == 5


class TestMain:

    def test_main(self, fuzzed_files, fasta):
        o = NamedTemporaryFile(delete=False)
        o.close()
        newref(fuzzed_files, o.name, reference=fasta.filename, binsize=100,
               n_bins=5)
        m = hashlib.md5()
        with open(o.name) as handle:
            for l in handle:
                m.update(l.encode('utf-8'))
        assert m.hexdigest() == "a719c83e318f75c296bcf41adb50593a"
        remove(o.name)



"""
setup.py
~~~~~~~~~~~~~
:copyright: (c) 2016-2019 Sander Bollen
:license: GPL-3.0
"""
from os.path import abspath, dirname, join

from setuptools import setup, find_packages


readme_file = join(abspath(dirname(__file__)), "README.md")
with open(readme_file) as desc_handle:
    long_desc = desc_handle.read()


setup(
    name="wisestork",
    version="0.1.1",
    description="Within-sample CNV calling",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    author="Sander Bollen",
    author_email="sander@sndrtj.eu",
    python_requires=">=3.5",
    license="GPLv3+",
    packages=find_packages(),
    url="https://github.com/sndrtj/wisestork",
    zip_safe=False,
    install_requires=[
        "numpy",
        "matplotlib",
        "statsmodels",
        "pysam",
        "pyfaidx",
        "biopython",
        "scipy",
        "click",
        "progressbar2"
    ],
    entry_points={
        "console_scripts": [
            "wisestork = wisestork.wisestork:main"
        ]
    },
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or "
        "later (GPLv3+)",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ]
)



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.

"""
wisestork.gc
~~~~~~~~~~
:copyright: (c) 2016-2019 Sander Bollen
:license: GPL-3.0
"""

from Bio.SeqUtils import GC


def get_gc_for_bin(fasta, chromosome, bin):
    """
    Get number of GC bases in a region
    :param fasta: an instance of pyfaidx.Fasta
    :param chromosome: chromosome name
    :param bin: Bin namedtuple
    :return: integer
    """
    perc = GC(fasta[chromosome][bin.start:bin.end].seq)
    dist = bin.end-bin.start
    return int((perc*dist)/100)


def get_n_per_bin(fasta, chromosome, bin):
    """
    Get number of N bases in a region
    :param fasta: an instance of pyfaidx.Fasta
    :param chromosome: chromosome name
    :param bin: Bin namedtuple
    :return: integer
    """
    return fasta[chromosome][bin.start:bin.end].seq.upper().count('N')



#    Copyright (C) 2016-2019  Sander Bollen
#
#    This file is part of wisestork
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see {http://www.gnu.org/licenses/}.
"""
wisestork.ztest
~~~~~~~~~~~~~

:copyright: (c) 2016-2019 Sander Bollen
:license: GPL-3.0
"""

import gzip
from math import isnan
import sys

import numpy as np
import progressbar

from .utils import BedLine, utf8


def get_z_score(bin, reference_bins):
    """
    Get z score of for this bin
    :param bin: bin
    :param reference_bins: reference bins
    :return: float
    """
    if len(reference_bins) == 0:
        return np.nan
    average = np.average([x.value for x in reference_bins])
    stddev = np.std([x.value for x in reference_bins])

    if stddev == 0:
        return np.nan
    Z = (bin.value - average) / stddev
    return Z


def ztest(input_path, output_path, database_path):
    """
    Calculate z scores from gzipped bed file and database bed file
    :param input_path: query bed file path
    :param output_path: output bed file path
    :param database_path: database file path
    :return: -
    """
    ohandle = open(output_path, "wb")
    refdict = build_reference_index(database_path)
    with gzip.open(input_path) as ihandle:
        bedlines = [BedLine.fromline(x) for x in ihandle]

    if not len(bedlines) == len(refdict):
        raise ValueError("Reference and query bed files "
                         "are of different size!")

    print("Calculating Z-scores")
    bar = progressbar.ProgressBar(max_value=len(bedlines))

    for i, x in enumerate(bedlines):
        bar.update(i)
        key = create_key(x)
        reference_bins = [bedlines[i] for i in refdict[key]]
        z = get_z_score(x, reference_bins)
        n = BedLine(x.chromosome.decode(), x.start, x.end, z)
        ohandle.write(bytes(n) + b"\n")
    bar.finish()
    ohandle.close()


def create_key(bedline):
    return b"|".join(map(utf8, [bedline.chromosome, bedline.start,
                                bedline.end]))


def build_reference_index(database_path):
    """
    Build a reference index from a database path.

    The index is a dictionary of type :
    {key: [list of indices]}

    where the key is a bed region, and the indices are indices of similar
    bins. Naturally, this means that query bed files _must_ be of
    _exactly_ the same size as the database, and the sorting order _must_
    also be identical.

    :param database_path:
    :return: index dictionary
    """
    print("Building index", file=sys.stderr)
    n = 0
    d = {}
    with gzip.open(database_path) as db_handle:
        for i, line in enumerate(db_handle):
            b = BedLine.fromline(line)
            key = create_key(b)
            d[key] = i
            n = i + 1
    refdict = {}
    with gzip.open(database_path) as db_handle2:
        bar = progressbar.ProgressBar(max_value=n)
        for o, l in enumerate(db_handle2):
            bar.update(o)
            b = BedLine.fromline(l)
            key2 = create_key(b)
            if isinstance(b.value, float) and isnan(b.value):
                refdict[key2] = []
            else:
                it = [BedLine.fromline(x, b",") for x in b.value.split(b"|")]
                keys = list(map(create_key, it))
                idxs = [d[x] for x in keys]
                refdict[key2] = idxs
        bar.finish()

    return refdict
